import six
import inspect
from grainy.core import (
    PermissionSet,
)

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

from .models import GrainyHandler
from .conf import PERM_CHOICES

def convert_flags(flags):
    """
    Converts string permission flags into integer permission flags
    
    Arguments:
        - flags <str>: one or more flags as they are defined in GRAINY_PERM_CHOICES
            
            For example: "crud" or "ru" or "r"
    
    Returns:
        - int
    """

    r = 0
    if not flags:
        return r

    if isinstance(flags, six.integer_types):
        return flags

    if not isinstance(flags, six.string_types):
        raise TypeError("`flags` needs to be a string or integer type")

    for f in flags:
        for f_i, name, f_s in PERM_CHOICES:
            if f_s == f:
                r = r | f_i
    return r


def namespace(target):

    """
    Convert `target` to permissioning namespace

    Arguments:
        - target <object|class|string>: if an object or class is passed here it 
            will be required to contain a `Grainy` meta class, otherwise a 
            TypeError will be raised.

    Returns:
        - string
    """

    if not target:
        return ""

    handler_class = getattr(target, "Grainy", None)

    if inspect.isclass(handler_class) and issubclass(handler_class, GrainyHandler):
        if inspect.isclass(target):
            return target.Grainy.namespace()
        return target.Grainy.namespace(instance=target)


    if isinstance(target, six.string_types):
        return target

    raise TypeError("`target` {} could not be convered to a permissioning namespace".format(target))

class Permissions(object):

    """
    A utility class that will allow you to perform permission checks
    for a user or group on cached permission sets
    """

    def __init__(self, obj):
        """
        Arguments:
            - obj <User|Group>
        """
        if not isinstance(obj, get_user_model()) and not isinstance(obj, Group):
            raise ValueError(
                "`obj` needs to be either of type `{}` or `Group`".format(
                    get_user_model().__class__.__name__
                )
            )
        self.obj = obj
        self.pset = PermissionSet()
        self.loaded = False
        self.load()

    def load(self, refresh=False):
        """
        Loads the permission set for the user or group specified in
        `self.obj` 

        Keyword Arguments:
            - refresh <bool>: if True, permission set will be reloaded if it
                has been loaded before
        """
        if not self.loaded or refresh:
            self.pset = self.obj.grainy_permissions.permission_set()
            if isinstance(self.obj, get_user_model()):
                # if we are loading permissions for user, we need
                # to also merge in permissions from any of the 
                # groups the user is a part of
                for group in self.obj.groups.all():
                    self.pset.update(
                        group.grainy_permissions.permission_set().permissions
                    )

    def check(self, target, permissions, explicit=False):
        """
        Check permissions for the specified target

        Arguments:
            - target <object|class|string>: check permissions to this object / namespace
            - permissions <int|string>: permission flags to check

        Keyword Arguments:
            - explicit <bool>: require explicit permissions to the complete target
                namespsace
        """
        return self.pset.check(namespace(target), convert_flags(permissions), explicit=explicit)

    def apply(self, data):
        return self.pset.apply(data)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-30 14:40
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_grainy.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0008_alter_user_username_max_length'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='GroupPermission',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('namespace', models.CharField(help_text="Permission namespace (A '.' delimited list of keys", max_length=255)),
                ('permission', django_grainy.fields.PermissionField(default=1)),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='grainy_permissions', to='auth.Group')),
            ],
            options={
                'verbose_name': 'Group Permission',
                'verbose_name_plural': 'Group Permissions',
                'base_manager_name': 'objects',
            },
        ),
        migrations.CreateModel(
            name='UserPermission',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('namespace', models.CharField(help_text="Permission namespace (A '.' delimited list of keys", max_length=255)),
                ('permission', django_grainy.fields.PermissionField(default=1)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='grainy_permissions', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'User Permission',
                'verbose_name_plural': 'User Permissions',
                'base_manager_name': 'objects',
            },
        ),
    ]



from django.conf import settings

def pytest_configure():
    settings.configure(
        SECRET_KEY="sekret",
        INSTALLED_APPS=[
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.admin',
            'django.contrib.sessions',
            'django_grainy',
            'django_grainy_test'
        ],
        DATABASE_ENGINE='django.db.backends.sqlite3',
        DATABASES={
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:',
            }
        },
        DEBUG=True
    )



from .util import UserTestCase
from django_grainy.models import UserPermission, GroupPermission
from django_grainy.fields import PermissionField

from grainy.const import (
    PERM_READ,
    PERM_UPDATE,
    PERM_CREATE,
    PERM_DELETE
)

class TestPermissionField(UserTestCase):

    @classmethod
    def setUpTestData(cls):
        UserTestCase.setUpTestData()

    def test_set(self):
        uperm = UserPermission()
        uperm.namespace = "test"
        uperm.user = self.users["user_a"]

        uperm.permission = PERM_READ | PERM_UPDATE
        uperm.full_clean()
        self.assertEqual(uperm.permission, PERM_READ | PERM_UPDATE)

        uperm.permission = "crud"
        uperm.full_clean()
        self.assertEqual(uperm.permission, PERM_READ | PERM_UPDATE | PERM_CREATE | PERM_DELETE)






from .models import GrainyHandler as _GrainyHandler

class grainy_model(object):

    def __init__(self, namespace=None):
        self.namespace = namespace

    def __call__(self, model):

        class Grainy(_GrainyHandler):
            pass

        Grainy.model = model

        if self.namespace is not None:
            namespace = self.namespace

            @classmethod
            def namespace_model(cls):
                return namespace.lower()
            Grainy.namespace_model = namespace_model

        model.Grainy = Grainy
        return model



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models

from django_grainy.decorators import grainy_model

# Create your models here.

"""
These are the models used during the django_grainy
unit tests. There is no need to ever install the "django_grainy_test"
app in your project
"""

class ModelBase(models.Model):
    class Meta(object):
        abstract = True

@grainy_model()
class ModelA(ModelBase):
    name = models.CharField(max_length=255)

@grainy_model(namespace="something.arbitrary")
class ModelB(ModelA):
    pass



import grainy.const
from django.utils.translation import ugettext_lazy as _

PERM_CHOICES_RW = [
    (grainy.const.PERM_READ, _("Read"), "r"),
    (grainy.const.PERM_WRITE, _("Write"), "w"),
]

PERM_CHOICES_CRUD = [
    (grainy.const.PERM_CREATE, _("Create"), "c"),
    (grainy.const.PERM_READ, _("Read"), "r"),
    (grainy.const.PERM_UPDATE, _("Update"), "u"),
    (grainy.const.PERM_DELETE, _("Delete"), "d"),
]



from .util import UserTestCase
from django.db import models
from django_grainy.models import (
    UserPermission,
    GroupPermission,
    GrainyHandler,
    PermissionSet
)
from django_grainy.decorators import grainy_model

from django_grainy_test.models import (
    ModelA,
    ModelB
)

from grainy.const import (
    PERM_READ,
    PERM_UPDATE,
    PERM_CREATE,
    PERM_DELETE,
)


class TestPermissionManager(UserTestCase):

    EXPECTED_PERMISSIONS_A = PermissionSet({
        "a" : PERM_READ,
        "b" : PERM_READ | PERM_UPDATE
    })

    EXPECTED_PERMISSIONS_A2 = PermissionSet({
        "a" : PERM_READ,
        "b" : PERM_READ | PERM_UPDATE,
        "c" : PERM_READ
    })

    EXPECTED_PERMISSIONS_B = PermissionSet({
        "c" : PERM_READ,
        "d" : PERM_CREATE
    })

    @classmethod
    def setUpTestData(cls):
        UserTestCase.setUpTestData()

        cls.users["user_a"].grainy_permissions.add_permission_set(
            cls.EXPECTED_PERMISSIONS_A
        )
        cls.users["user_b"].grainy_permissions.add_permission_set(
            cls.EXPECTED_PERMISSIONS_B
        )


    def test_permission_set_from_query(self):
        pset = self.users["user_a"].grainy_permissions.permission_set()
        self.assertEqual(self.EXPECTED_PERMISSIONS_A.permissions, pset.permissions)

        pset = self.users["user_b"].grainy_permissions.permission_set()
        self.assertEqual(self.EXPECTED_PERMISSIONS_B.permissions, pset.permissions)

    def test_permission_set_upates(self):
        self.users["user_a"].grainy_permissions.add_permission_set(
            PermissionSet({"c" : PERM_READ})
        )
        pset = self.users["user_a"].grainy_permissions.permission_set()
        self.assertEqual(self.EXPECTED_PERMISSIONS_A2.permissions, pset.permissions)

        self.users["user_a"].grainy_permissions.add_permission_set(
            PermissionSet({"c" : PERM_READ}),
            clear=True
        )
        pset = self.users["user_a"].grainy_permissions.permission_set()
        self.assertEqual(PermissionSet({"c": PERM_READ}).permissions, pset.permissions)












from .util import UserTestCase
from django.db import models
from django_grainy.models import (
    UserPermission,
    GroupPermission,
    GrainyHandler,
    PermissionSet
)
from django_grainy.decorators import grainy_model

from django_grainy_test.models import (
    ModelA,
    ModelB
)

from django_grainy.util import (
    namespace,
    convert_flags,
    Permissions
)

from grainy.const import (
    PERM_READ,
    PERM_UPDATE,
    PERM_CREATE,
    PERM_DELETE,
)

class TestPermissions(UserTestCase):

    EXPECTED_PERMISSIONS_A = PermissionSet({
        ModelA.Grainy.namespace() : PERM_READ,
        ModelB.Grainy.namespace() : PERM_READ | PERM_UPDATE
    })

    GROUP_PERMISSIONS_A = PermissionSet({
        "secret.group" : PERM_READ
    })


    @classmethod
    def setUpTestData(cls):
        UserTestCase.setUpTestData()

        cls.users["user_a"].grainy_permissions.add_permission_set(
            cls.EXPECTED_PERMISSIONS_A
        )

        cls.groups["group_a"].grainy_permissions.add_permission_set(
            cls.GROUP_PERMISSIONS_A
        )

    def test_namespace(self):
        """
        test django_grainy.util.namespace
        """

        self.assertEqual(namespace(ModelA), "django_grainy_test.modela")
        self.assertEqual(namespace(ModelA()), "django_grainy_test.modela.none")
        self.assertEqual(namespace("a.b.c"), "a.b.c")
        self.assertEqual(namespace(None), "")
        with self.assertRaises(TypeError):
            namespace(object())

    def test_convert_flags(self):
        """
        test django_grainy.util.convert_flags
        """

        self.assertEqual(convert_flags("c"), PERM_CREATE)
        self.assertEqual(convert_flags("cr"), PERM_CREATE | PERM_READ)
        self.assertEqual(convert_flags("cru"), PERM_CREATE | PERM_READ | PERM_UPDATE)
        self.assertEqual(convert_flags("crud"), PERM_CREATE | PERM_READ | PERM_UPDATE | PERM_DELETE)
        self.assertEqual(convert_flags("xyz"), 0)
        self.assertEqual(convert_flags(None), 0)

        self.assertEqual(convert_flags(convert_flags("c")), PERM_CREATE)
        self.assertEqual(convert_flags(convert_flags("cr")), PERM_CREATE | PERM_READ)
        self.assertEqual(convert_flags(convert_flags("cru")), PERM_CREATE | PERM_READ | PERM_UPDATE)
        self.assertEqual(convert_flags(convert_flags("crud")), PERM_CREATE | PERM_READ | PERM_UPDATE | PERM_DELETE)

        with self.assertRaises(TypeError):
            convert_flags(object())


    def test_permissions_init(self):
        """
        test django.grainy.util.Permissions.__init__
        """
        user = self.users["user_a"]
        group = self.groups["group_a"]

        self.assertEqual(Permissions(user).obj, user)
        self.assertEqual(Permissions(group).obj, group)
        with self.assertRaises(ValueError):
            Permissions(object())

    def test_permissions_check(self):
        """
        test django.grainy.util.Permissions.check
        """
        user = self.users["user_a"]

        perms = Permissions(user)
        self.assertTrue(perms.check(ModelA, PERM_READ))
        self.assertFalse(perms.check(ModelA, PERM_UPDATE))
        self.assertTrue(perms.check(ModelB, PERM_READ))
        self.assertTrue(perms.check(ModelB, PERM_UPDATE))
        self.assertTrue(perms.check(ModelA(), PERM_READ))
        self.assertFalse(perms.check(ModelA(), PERM_UPDATE))
        self.assertTrue(perms.check(ModelB(), PERM_READ))
        self.assertTrue(perms.check(ModelB(), PERM_UPDATE))
        self.assertTrue(perms.check("secret.group", PERM_READ))
        self.assertFalse(perms.check("secret", PERM_READ))










# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-27 14:25
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ModelA',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ModelB',
            fields=[
                ('modela_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='django_grainy_test.ModelA')),
            ],
            options={
                'abstract': False,
            },
            bases=('django_grainy_test.modela',),
        ),
    ]



from django.test import TestCase
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

class UserTestCase(TestCase):
    setup_users = ["user_a", "user_b", "user_c"]
    setup_groups = {
        "group_a" : ["user_a"],
        "group_b" : ["user_b"],
        "group_c" : ["user_c"]
    }
    @classmethod
    def setUpTestData(cls):
        # create users
        cls.users = dict([
            (k, get_user_model().objects.create_user(k, "{}@example.com".format(k), k))
            for k in cls.setup_users
        ])

        # create usergroups
        cls.groups = dict([
            (k, Group.objects.create(name=k)) for k in cls.setup_groups
        ])

        # add users to groups
        for group in cls.groups.values():
            for username in cls.setup_groups.get(group.name):
                group.user_set.add(cls.users.get(username))






# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.utils.translation import ugettext_lazy as _

from grainy.const import PERM_READ
from grainy.core import PermissionSet

from .fields import PermissionField
from .conf import PERM_CHOICES

class PermissionQuerySet(models.QuerySet):
    """
    Queryset that offers additional utilities for UserPermission and
    GroupPermission queries
    """

    def permission_set(self):
        """
        Builds and returns a grainy.PermissionSet object from all
        the rows in the query

        Returns:
            - grainy.PermissionSet 
        """
        pset = PermissionSet()
        for row in self:
            pset[row.namespace] = row.permission
        return pset

class PermissionManager(models.Manager):
    """
    Object Manager that offers additional utilities for managing
    UserPermission and GroupPermission objects
    """

    def get_queryset(self):
        return PermissionQuerySet(self.model, using=self._db)

    def add_permission_set(self, pset, clear=False):
        """
        Add all permissions specified in a PermissionSet

        Arguments:
            - pset <grainy.PermissionSet>

        Keyword Arguments:
            - clear <bool>: if true, clear all existing permissions before
                adding the new set.
        """

        if clear:
            _pset = PermissionSet()
        else:
            _pset = self.permission_set()

        self.get_queryset().all().delete()

        for namespace, permission in pset.permissions.items():
            _pset[namespace] = permission

        for namespace, permission in _pset.permissions.items():
            self.create(namespace=namespace, permission=permission.value)

    def permission_set(self):
        """
        Return grainy.PermissionSet instance from all rows returned
        from get_queryset()

        Returns:
            - grainy.PermissionSet
        """
        return self.get_queryset().all().permission_set()

# Create your models here.

class Permission(models.Model):
    class Meta(object):
        abstract = True

    namespace = models.CharField(
        max_length=255,
        help_text=_("Permission namespace (A '.' delimited list of keys")
    )
    permission = PermissionField(
        default=PERM_READ
    )

    def __unicode__(self):
        return u"{}: {}".format(self.namespace, self.permission)

class UserPermission(Permission):

    """
    Describes permission for a user
    """
    class Meta(object):
        verbose_name = _("User Permission")
        verbose_name_plural = _("User Permissions")
        base_manager_name = "objects"

    user = models.ForeignKey(get_user_model(), related_name="grainy_permissions", on_delete=models.CASCADE)
    objects = PermissionManager()

class GroupPermission(Permission):
    """
    Describes permission for a user group
    """
    class Meta(object):
        verbose_name = _("Group Permission")
        verbose_name_plural = _("Group Permissions")
        base_manager_name = "objects"

    group = models.ForeignKey(Group, related_name="grainy_permissions", on_delete=models.CASCADE)
    objects = PermissionManager()


class GrainyHandler(object):

    """
    The base class to use for the Grainy Meta class to put inside
    Models that you want to use grainy permissions on
    """

    # the model handled
    model = None

    @classmethod
    def namespace_instance(cls, instance):
        """
        Returns the permissioning namespace for the model instance
        passed.

        Arguments:
            - instance (models.Model): model instance

        Returns:
            - unicode
        """
        return u"{}.{}".format(
            cls.namespace_model(),
            instance.id
        ).lower()

    @classmethod
    def namespace_model(cls):
        """
        Returns the permissioning namespace for the model specified
        in cls.model

        Returns:
            - unicode
        """
        return u"{}.{}".format(
            cls.model._meta.app_label,
            cls.model._meta.object_name
        ).lower()

    @classmethod
    def namespace(cls, instance=None):
        """
        Returns the permissioning namespace of the model class mangaged
        by this handler, or an instance of said model if it is specified.

        Keyword Arguments:
            - instance <models.Model>: model instance, if specified 
                the permissioning namespace returned will be fore this instance.

        Returns:
            - unicode
        """
        if instance:
            return cls.namespace_instance(instance)
        return cls.namespace_model()




import six
from django.db import models

from .conf import PERM_CHOICES

class PermissionField(models.IntegerField):

    def to_python(self, value):
        # if a string is passed it should be parsed
        # for string flags (for example 'c', 'r', 'u' and 'd')
        # as specified in the PERM_CHOICES setting and
        # convert to bitmask
        if isinstance(value, six.string_types):
            _value = 0
            for flag, name, str_flag in PERM_CHOICES:
                if str_flag in value:
                    _value |= flag
            value = _value

        elif value is None:
            return 0

        return value





from .util import UserTestCase
from django.db import models
from django_grainy.models import UserPermission, GroupPermission, GrainyHandler
from django_grainy.decorators import grainy_model

from django_grainy_test.models import (
    ModelA,
    ModelB
)

from grainy.const import (
    PERM_READ,
    PERM_UPDATE,
    PERM_CREATE,
    PERM_DELETE
)


class TestGrainyHandler(UserTestCase):

    @classmethod
    def setUpTestData(cls):
        UserTestCase.setUpTestData()

    def test_grainy_model_decorator(self):
        a = ModelA.objects.create(name="A1")
        self.assertEqual(a.Grainy.namespace(a), "django_grainy_test.modela.{}".format(a.id))
        b = ModelB.objects.create(name="B1")
        self.assertEqual(b.Grainy.namespace(b), "something.arbitrary.{}".format(b.id))

